#!/usr/bin/env python

from __future__ import division

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.']

from math   import ceil
import traceback
import argparse
import random
import time
import numpy as np
from collections import defaultdict
from csv import DictWriter
from util.stats import proportion_ci

from multiprocessing import Pool, cpu_count
from schedcat.util.storage import storage
from generator.generator import DesignPointGenerator

from distributions import *

from schedcat.model.LLC_params import *
from schedcat.generator.mcx_tasks import TaskGeneratorMCX
from LLCOptimizer import *
from LLCCanvas import LLCCanvas
from DAGGenerator import *
import schedcat.model.ohc as ohc
import loadohs
import graph

from BufferSizer import *

from sched_check import *

import multiprocessing
# We must import this explicitly, it is not imported by the top-level
# multiprocessing module.
import multiprocessing.pool
from random import randint

MIN_SAMPLES = 100
MAX_SAMPLES = 2000 #2000
MAX_CI      = 0.05 #.05
CONFIDENCE  = 0.95 #.95

topdir = "~/LLC_Code/MCXSchedStudy/"

def make_mcx_task_set(dp):
    tgMCX = TaskGeneratorMCX(
                             period = NAMED_PERIODS[dp.period],
                             util = NAMED_UTILIZATIONS[dp.task_util],
                             util_dist = NAMED_UTIL_DISTS[dp.util_dist],
                             mem_util = NAMED_MEMORY[dp.memory_util],
                             avg_percent = AVG_PERCENT,
                             pfuncs = P)
    ts = tgMCX.make_task_set(max_util = dp.sys_util)
    
    ts.m = CORES
        
    ts.scheme = bypass_scheme()
    
    #mem = uniform(0.35, 0.45)
    libn = {'A' : uniform_int(4, 5), 'B' : uniform_int(4, 5), 'C' : uniform_int(6, 10)}
    
    drambump = NAMED_DYN_REDUCTIONS[dp.dyn_red]
    
    #private page: static schemes
    #private page: dynamic schemes 
    #50-300 difference
    #each core/C: dynamic library instructions
    
    dramdist = NAMED_MEM_FOOTPRINTS[dp.dram]
    
    for tsk in ts:
        tsk.ddram = ceil(dramdist[tsk.level]())
        statbump = drambump[tsk.level]()
        
        tsk_libn = libn[tsk.level]()-len(NAMED_LIBRARIES_COMM.keys())
        libs = NAMED_LIBRARIES_SPEC[tsk.level]
        maxn = len(libs)
        tsk_libs = list(libs.keys())

        for i in range(0, maxn-tsk_libn):
            del tsk_libs[uniform_int(0, maxn-i-1)()]
        
        tsk.libs = defaultdict()
        for lib in tsk_libs:
            tsk.libs[lib] = NAMED_LIBRARIES[lib]
        
        for lib, size in NAMED_LIBRARIES_COMM.iteritems():
            tsk.libs[lib] = size
        
        libsum = 0
        for lib, size in tsk.libs.iteritems():
            libsum += size
        statbump = min(statbump, 0.9*libsum)
        
        tsk.sdram = ceil(tsk.ddram + statbump)
        
        tsk.wss = min(tsk.sdram, tsk.wss)
    
    return ts

def complete(results, n):
    if n < MIN_SAMPLES:
        return False
    elif n > MAX_SAMPLES:
        return True
    else:
        for key in results:
            if proportion_ci(results[key], n, CONFIDENCE) > MAX_CI:
                return False
        return True

def update_mean(old_mean, n, new_sample):
    return (old_mean*n + new_sample)/(n+1)

def timeout(func, ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
    '''This function will spwan a thread and run the given function using the args, kwargs and 
    return the given default value if the timeout_duration is exceeded 
    ''' 
    import threading
    class InterruptableThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self.result = 0
        def run(self):
            try:
#                 print "started with: "
                self.result = func(ts_mcx, vartype, csch, olap = olap, debug = debug, noIso = noIso)
#                 print "finished with", self.result
            except:
                self.result = 0
                
    it = InterruptableThread()
    it.start()
    it.join(10)
    
    if it.result == 0:
        print "didn't finish"
        return None
    else:
#         print "returning:", it.result
        return it.result

core_pgs_AB = 0
pgs_C = 0
bank_pgs = 128*256

def set_mem_non_int(dp):
    global core_pgs_AB
    global pgs_C
    mode_cap = NAMED_CAPACITIES[dp.capacity]
    core_pgs_AB = bank_pgs*0.25*mode_cap
    pgs_C = 2.25*bank_pgs*mode_cap

#OS: 30% of each bank.
def set_mem_int(dp):
    global core_pgs_AB
    global pgs_C
    mode_cap = NAMED_CAPACITIES[dp.capacity]
    core_pgs_AB = bank_pgs*0.75*mode_cap
    pgs_C = 4*0.75*bank_pgs*mode_cap

def mem_overflow_lib_obl(ts):
    ts_AB = ts.level('A')
    
    ts_AB.extend(ts.level('B'))
    
#     print "  library oblivious"
    for core in range(0, 4):
#         print "    AB, core", core, ":", '{0:.0f}'.format(ts_AB.core(core).mem_lib_obl()), "out of", core_pgs_AB, "pages"
        if ts_AB.core(core).mem_lib_obl() > core_pgs_AB:
            return True
    
#     print "    C :", '{0:.0f}'.format(ts.level('C').mem_lib_obl()), "out of", pgs_C, "pages"
    if ts.level('C').mem_lib_obl() > pgs_C:
        return True
    
    return False

def mem_overflow_static(ts):
    ts_AB = ts.level('A')
    
    ts_AB.extend(ts.level('B'))
    
#     print "  static linking"
    for core in range(0, 4):
#         print "    AB, core", core, ":", '{0:.0f}'.format(ts_AB.core(core).mem_static()), "out of", core_pgs_AB, "pages"
        if ts_AB.core(core).mem_static() > core_pgs_AB:
            return True
    
#     print "    C :", '{0:.0f}'.format(ts.level('C').mem_static()), "out of", pgs_C, "pages"
    if ts.level('C').mem_static() > pgs_C:
        return True
    
    return False

def mem_overflow_dynam(ts):
    
    ts_AB = ts.level('A')
    
    ts_AB.extend(ts.level('B'))
    
#     print "  dynamic linking"
    for core in range(0, 4):
#         print "    AB, core", core, ":", '{0:.0f}'.format(ts_AB.core(core).mem_dynam()), "out of", core_pgs_AB, "pages"
        if ts_AB.core(core).mem_dynam() > core_pgs_AB:
            return True
    
#     print "    C :", '{0:.0f}'.format(ts.level('C').mem_dynam()), "out of", pgs_C, "pages"
    if ts.level('C').mem_dynam() > pgs_C:
        return True
        
    return False

def process_dp(dp):
    
    sched_res = defaultdict(float)
    count = defaultdict(int)
    
    n = 0
    
    while not complete(sched_res, n):
        ts = make_mcx_task_set(dp)
        
        unmnwts = []
        loadwtsAB = []
        loadwtsC = []
    
        for tsk in ts:
            unmnwts.append(unmanwt())
            loadwtsAB.append(loadedinfAB())
        
        for tsk in ts.level('C'):
                loadwtsC.append(loadedinfC())
    
        ts.useCshared()
        ts.setLLCAlloc(bypass_scheme())
        ohc.SetOheads(ts, 'mc2') 
    
        #---------inflate for loads in MC-Cshared------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                    
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
    
        ts.coreAssign(useMC2 = True)
    
        set_mem_non_int(dp)
        niAsMemStatSched = not mem_overflow_static(ts)
        niAsMemDynaSched = not mem_overflow_dynam(ts)
        niAsMemLbObSched = not mem_overflow_lib_obl(ts)
#         print
    
        ts.setLLCAlloc(bypass_scheme())
        
        psNonI = LPLLCWays(ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psNonI == None:
            psNonI = LPLLCWays(ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
        schedNonI = psNonI != None 
        
        #-----------------Interleaved-----------------#
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
                        for crt in {'A', 'B', 'C'}:
                            tsk.cost[crt][s][w] *= unmnwts[idx]
        
        set_mem_int(dp)
        interAsMemStatSched = not mem_overflow_static(ts)
        interAsMemDynaSched = not mem_overflow_dynam(ts)
        interAsMemLbObSched = not mem_overflow_lib_obl(ts)
#         print
        ts.setLLCAlloc(inter_scheme())
        
        psInter = LPLLCWaysInter(ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psInter == None:
            psInter = LPLLCWaysInter(ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
        schedInter = psInter != None
        
        # for core in range(0,4):
#             print str(core)+":", psInter.A[core].S, psInter.A[core].W
#         print psInter.C.S, psInter.C.W
#         print
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
                        for crt in {'A', 'B', 'C'}:
                            tsk.cost[crt][s][w] /= unmnwts[idx]
        ###UEDF###
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsAB[idx]    
                    
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsC[idx] 
    
        ts.useCiso()
    
        for tsk in ts:
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] = tsk.cost['A'][s][w]   
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] /= unmnwts[idx]
    
        for tsk in ts:
            tsk.core = 0
            tsk.level = 'B'
    
        ts.setLLCAlloc(uni_scheme())
        ohc.SetOheads(ts, 'pedf')
        mx_wss_all = max([tsk.wss for tsk in ts])
        mx_wss = {'B' : [mx_wss_all, mx_wss_all, mx_wss_all, mx_wss_all],
                  'C' : 0}
        ohc.CPMD = 30
        schedUEDF = ts.utilization('B', approximate = False, inflated = True, c_taskset = [], minAperiods = [0, 0, 0, 0], mx_wss = mx_wss) <= 1
    
        # try:
#             assert (not schedCLinfm) or schedIdeal
#         except AssertionError:
#             _, _, tb = sys.exc_info()
#             tb_info = traceback.extract_tb(tb)
#             _, _, _, text = tb_info[-1]
# 
#             print "Assertion error", text
#             print "dp: ", dp
#             print "utils:", schedCLdata, schedIdealdata 
#             raise
        
        sched_res['U-EDF']   = update_mean(sched_res['U-EDF'], n, schedUEDF)
        
        sched_res['stat-NI']    = update_mean(sched_res['stat-NI']   , n, schedNonI  and niAsMemStatSched)
        sched_res['dyna-NI']    = update_mean(sched_res['dyna-NI']   , n, schedNonI  and niAsMemDynaSched)
        sched_res['lbob-NI']    = update_mean(sched_res['lbob-NI']   , n, schedNonI  and niAsMemLbObSched)
        sched_res['infm-NI']    = update_mean(sched_res['infm-NI']   , n, schedNonI )
        
        sched_res['stat-inter'] = update_mean(sched_res['stat-inter'], n, schedInter and interAsMemStatSched)
        sched_res['dyna-inter'] = update_mean(sched_res['dyna-inter'], n, schedInter and interAsMemDynaSched)
        sched_res['lbob-inter'] = update_mean(sched_res['lbob-inter'], n, schedInter and interAsMemLbObSched)
        sched_res['infm-inter'] = update_mean(sched_res['infm-inter'], n, schedInter)
        
        count['Count'] = n+1

        n+=1
    
    return dict(dp.items() + sched_res.items()+count.items())
    

def myrange(start, end, inc):
    return np.arange(start, end+inc, inc)
    
def valid(dp):
    return True

def thread_tests(args, outs, design_points):
    pool = Pool(processes = args.processors)
    try:
        for i, row in enumerate(pool.imap_unordered(process_dp, design_points)):
            if not row:
                continue
            sys.stderr.write('\rdone {0:%}'.format(i/len(design_points)))
#             print '\rdone {0:%}'.format(i/len(design_points))
#             print row
            title = titlefor(row)
            outs[title].writerow(row)
        pool.close()
    except Exception as e:
        print "hit an exception"
        print e
        pool.terminate()
        pool.join()

def titlefor(headers):
    return headers['util_dist'] + "--" + headers['period'] + "--" + headers['task_util'] + "--" + headers['memory_util'] + "--" + headers['capacity'] + "--" + headers['dram'] + "--" + headers['dyn_red']

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help = "Number of processors to execute on")

    args = parser.parse_args()
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.capacity = NAMED_CAPACITIES.keys()
#     exp.stat_opt = NAMED_STAT_OPT.keys()
    exp.dram = NAMED_MEM_FOOTPRINTS.keys()
    exp.dyn_red = NAMED_DYN_REDUCTIONS.keys()
    exp.sys_util = myrange(0.1, 9.1, 0.2)
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
    # random.shuffle(design_points)
    
    loadohs.load("../oheads/")
    
    print "Total design points: ", len(design_points)
    # if(len(os.listdir(os.getcwd() + "/results/")) > 0):
        # print "already have results"
        # exit()
    
    outs = dict()
    for dp in design_points:
        title = titlefor(dp)
        if not title in outs:
            outs[title] = DictWriter(open("../results/" + title, 'w'), exp.keys()+['U-EDF', 'stat-NI', 'dyna-NI', 'lbob-NI', 'infm-NI', 'stat-inter', 'dyna-inter', 'lbob-inter', 'infm-inter', 'Count'])
            outs[title].writeheader()

#     for i, t in enumerate(design_points):
#         process_dp(design_points[i])

    start = time.time()
    print "start:", start
    thread_tests(args, outs, design_points)
    end = time.time()
    print ""
    print "end:", end
    
    print "time:", end-start
    print "in hours:", (end-start)/3600.0
    
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help = "Number of processors to execute on")

    args = parser.parse_args()
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.capacity = NAMED_CAPACITIES.keys()
#     exp.stat_opt = NAMED_STAT_OPT.keys()
    exp.dram = NAMED_MEM_FOOTPRINTS.keys()
    exp.dyn_red = NAMED_DYN_REDUCTIONS.keys()
    exp.sys_util = myrange(0.1, 9.1, 0.2)
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
    # random.shuffle(design_points)
    
    loadohs.load("../oheads/")
    
    print "Total design points: ", len(design_points)
    # if(len(os.listdir(os.getcwd() + "/results/")) > 0):
        # print "already have results"
        # exit()
    
    outs = dict()
    for dp in design_points:
        title = titlefor(dp)
        if not title in outs:
            outs[title] = DictWriter(open("../results/" + title, 'w'), exp.keys()+['U-EDF', 'stat-NI', 'dyna-NI', 'lbob-NI', 'infm-NI', 'stat-inter', 'dyna-inter', 'lbob-inter', 'infm-inter', 'Count'])
            outs[title].writeheader()

#     for i, t in enumerate(design_points):
#         process_dp(design_points[i])

    start = time.time()
    print "start:", start
    thread_tests(args, outs, design_points)
    end = time.time()
    print ""
    print "end:", end
    
    print "time:", end-start
    print "in hours:", (end-start)/3600.0

if __name__ == '__main__':
    main()